/* dir_name_max(dir) does the same as pathconf(dir, _PC_NAME_MAX) */

#ifdef PATHCONF_MISSING

#include <sys/types.h>

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif /* HAVE_LIMITS_H */

#ifdef HAVE_DIRENT_H
#include <dirent.h>
#endif /* HAVE_DIRENT_H */

#ifdef HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif /* HAVE_SYS_DIR_H */

#ifndef NAME_MAX
#ifdef MAXNAMLEN
#define NAME_MAX MAXNAMLEN
#else /* !MAXNAMLEN */
#ifdef MAXNAMELEN
#define NAME_MAX MAXNAMELEN
#else /* !MAXNAMELEN */
#define NAME_MAX 14
#endif /* !MAXNAMELEN */
#endif /* !MAXNAMLEN */
#endif /* !NAME_MAX */

long dir_name_max(dir)
     char *dir;
{
  return NAME_MAX;
}

#else /* not PATHCONF_MISSING */

#include <unistd.h>

long dir_name_max(dir)
     char *dir;
{
  return pathconf(dir, _PC_NAME_MAX);
}

#endif /* not PATHCONF_MISSING */
